@@ -11,6 +11,7 @@
 #include "config-konsole.h"
 
 // Qt
+#include <QFile>
 #include <QTextStream>
 
 // Konsole decoders
@@ -70,7 +71,10 @@ Screen::Screen(int lines, int columns)
     , _cuY(0)
     , _currentForeground(CharacterColor())
     , _currentBackground(CharacterColor())
-    , _currentRendition(DEFAULT_RENDITION)
+    , _currentRendition({DEFAULT_RENDITION})
+    , _ulColorQueueStart(0)
+    , _ulColorQueueEnd(0)
+    , _currentULColor(0)
     , _topMargin(0)
     , _bottomMargin(0)
     , _replMode(REPL_None)
@@ -83,12 +87,12 @@ Screen::Screen(int lines, int columns)
     , _blockSelectionMode(false)
     , _effectiveForeground(CharacterColor())
     , _effectiveBackground(CharacterColor())
-    , _effectiveRendition(DEFAULT_RENDITION)
+    , _effectiveRendition({DEFAULT_RENDITION})
     , _lastPos(-1)
     , _lastDrawnChar(0)
     , _escapeSequenceUrlExtractor(nullptr)
 {
-    std::fill(_lineProperties.begin(), _lineProperties.end(), LINE_DEFAULT);
+    std::fill(_lineProperties.begin(), _lineProperties.end(), LineProperty());
 
     _graphicsPlacements = std::vector<std::unique_ptr<TerminalGraphicsPlacement_t>>();
     _hasGraphics = false;
@@ -172,6 +176,117 @@ void Screen::cursorRight(int n)
     _cuX = qMin(getScreenLineColumns(_cuY) - 1, _cuX + n);
 }
 
+void Screen::initSelCursor()
+{
+    _selCuX = _cuX;
+    _selCuY = _cuY;
+}
+
+int Screen::selCursorUp(int n)
+{
+    if (n == 0) {
+        // Half page
+        n = _lines / 2;
+    } else if (n == -1) {
+        // Full page
+        n = _lines;
+    } else if (n == -2) {
+        // First line
+        n = _selCuY + _history->getLines();
+    }
+    _selCuY = qMax(-_history->getLines(), _selCuY - n);
+    return _selCuY;
+}
+
+int Screen::selCursorDown(int n)
+{
+    if (n == 0) {
+        // Half page
+        n = _lines / 2;
+    } else if (n == -1) {
+        // Full page
+        n = _lines;
+    } else if (n == -2) {
+        // Last line
+        n = _lines - 1 - _selCuY;
+    }
+    _selCuY = qMin(_lines - 1, _selCuY + n);
+    return _selCuY;
+}
+
+int Screen::selCursorLeft(int n)
+{
+    if (n == 0) {
+        // Home
+        n = _selCuX;
+    }
+    if (_selCuX >= n) {
+        _selCuX -= n;
+    } else {
+        if (_selCuY > -_history->getLines()) {
+            _selCuY -= 1;
+            _selCuX = qMax(_columns - n + _selCuX, 0);
+        } else {
+            _selCuX = 0;
+        }
+    }
+    return _selCuY;
+}
+
+int Screen::selCursorRight(int n)
+{
+    if (n == 0) {
+        // End
+        n = _columns - _selCuX - 1;
+    }
+    if (_selCuX + n < _columns) {
+        _selCuX += n;
+    } else {
+        if (_selCuY < _lines - 1) {
+            _selCuY += 1;
+            _selCuX = qMin(n + _selCuX - _columns, _columns - 1);
+        } else {
+            _selCuX = _columns - 1;
+        }
+    }
+    return _selCuY;
+}
+
+int Screen::selSetSelectionStart(int mode)
+{
+    // mode: 0 = character selection
+    //       1 = line selection
+    int x = _selCuX;
+    if (mode == 1) {
+        x = 0;
+    }
+    setSelectionStart(x, _selCuY + _history->getLines(), false);
+    return 0;
+}
+
+int Screen::selSetSelectionEnd(int mode)
+{
+    int y = _selCuY + _history->getLines();
+    int x = _selCuX;
+    if (mode == 1) {
+        int l = _selBegin / _columns;
+        if (y < l) {
+            if (_selBegin % _columns == 0) {
+                setSelectionStart(_columns - 1, l, false);
+            }
+            x = 0;
+        } else {
+            x = _columns - 1;
+            if (_selBegin % _columns != 0) {
+                setSelectionStart(0, l, false);
+            }
+        }
+    }
+    setSelectionEnd(x, y, false);
+    Q_EMIT _currentTerminalDisplay->screenWindow()->selectionChanged();
+    return 0;
+}
+
 void Screen::setMargins(int top, int bot)
 //=STBM
 {
@@ -225,7 +340,7 @@ void Screen::reverseIndex()
 void Screen::nextLine()
 //=NEL
 {
-    _lineProperties[_cuY] = SetLineLength(_lineProperties[_cuY], _cuX);
+    _lineProperties[_cuY].length = _cuX;
     toStartOfLine();
     index();
 }
@@ -286,7 +401,7 @@ void Screen::deleteChars(int n)
     _screenLines[_cuY].remove(_cuX, n);
 
     // Append space(s) with current attributes
-    Character spaceWithCurrentAttrs(' ', _effectiveForeground, _effectiveBackground, _effectiveRendition, 0);
+    Character spaceWithCurrentAttrs(' ', _effectiveForeground, _effectiveBackground, _effectiveRendition.all, 0);
 
     for (int i = 0; i < n; ++i) {
         _screenLines[_cuY].append(spaceWithCurrentAttrs);
@@ -517,14 +632,14 @@ void Screen::resizeImage(int new_lines, int new_columns)
             auto sessionController = _currentTerminalDisplay->sessionController();
             auto terminal = sessionController->session()->foregroundProcessName();
             if (terminal == QLatin1String("zsh")) {
-                while (cursorLine + cursorLineCorrection > 0 && (_lineProperties.at(cursorLine + cursorLineCorrection) & LINE_PROMPT_START) == 0) {
+                while (cursorLine + cursorLineCorrection > 0 && (_lineProperties.at(cursorLine + cursorLineCorrection).flags.f.prompt_start) == 0) {
                     --cursorLineCorrection;
                 }
-                if (cursorLine + cursorLineCorrection > 0 && (_lineProperties.at(cursorLine + cursorLineCorrection) & LINE_PROMPT_START) != 0) {
-                    _lineProperties[cursorLine + cursorLineCorrection - 1] &= ~LINE_WRAPPED;
+                if (cursorLine + cursorLineCorrection > 0 && (_lineProperties.at(cursorLine + cursorLineCorrection).flags.f.prompt_start) != 0) {
+                    _lineProperties[cursorLine + cursorLineCorrection - 1].flags.f.wrapped = 0;
                 } else {
                     cursorLineCorrection = 0;
-                    while (cursorLine + cursorLineCorrection > 0 && (_lineProperties.at(cursorLine + cursorLineCorrection - 1) & LINE_WRAPPED) != 0) {
+                    while (cursorLine + cursorLineCorrection > 0 && (_lineProperties.at(cursorLine + cursorLineCorrection - 1).flags.f.wrapped) != 0) {
                         --cursorLineCorrection;
                     }
                 }
@@ -535,7 +650,7 @@ void Screen::resizeImage(int new_lines, int new_columns)
         int currentPos = 0;
         while (currentPos < (cursorLine + cursorLineCorrection) && currentPos < (int)_screenLines.size() - 1) {
             // Join wrapped line in current position
-            if ((_lineProperties.at(currentPos) & LINE_WRAPPED) != 0) {
+            if ((_lineProperties.at(currentPos).flags.f.wrapped) != 0) {
                 _screenLines[currentPos].append(_screenLines.at(currentPos + 1));
                 _screenLines.erase(_screenLines.begin() + currentPos + 1);
                 _lineProperties.erase(_lineProperties.begin() + currentPos);
@@ -551,12 +666,13 @@ void Screen::resizeImage(int new_lines, int new_columns)
             }
 
             // If need to move to line below, copy from the current line, to the next one.
-            if (lineSize > new_columns && !(_lineProperties.at(currentPos) & (LINE_DOUBLEHEIGHT_BOTTOM | LINE_DOUBLEHEIGHT_TOP))) {
+            if (lineSize > new_columns
+                && !(_lineProperties.at(currentPos).flags.f.doubleheight_bottom | _lineProperties.at(currentPos).flags.f.doubleheight_top)) {
                 auto values = _screenLines.at(currentPos).mid(new_columns);
                 _screenLines[currentPos].resize(new_columns);
                 _lineProperties.insert(_lineProperties.begin() + currentPos + 1, _lineProperties.at(currentPos));
                 _screenLines.insert(_screenLines.begin() + currentPos + 1, std::move(values));
-                _lineProperties[currentPos] |= LINE_WRAPPED;
+                _lineProperties[currentPos].flags.f.wrapped = 1;
                 ++cursorLine;
                 scrollPlacements(-1, currentPos);
             }
@@ -590,7 +706,7 @@ void Screen::resizeImage(int new_lines, int new_columns)
 
     _lineProperties.resize(new_lines + 1);
     if (_lineProperties.size() > _screenLines.size()) {
-        std::fill(_lineProperties.begin() + _screenLines.size(), _lineProperties.end(), LINE_DEFAULT);
+        std::fill(_lineProperties.begin() + _screenLines.size(), _lineProperties.end(), LineProperty());
     }
     _screenLines.resize(new_lines + 1);
 
@@ -646,7 +762,7 @@ void Screen::reverseRendition(Character &p) const
 void Screen::updateEffectiveRendition()
 {
     _effectiveRendition = _currentRendition;
-    if ((_currentRendition & RE_REVERSE) != 0) {
+    if ((_currentRendition.f.reverse) != 0) {
         _effectiveForeground = _currentBackground;
         _effectiveBackground = _currentForeground;
     } else {
@@ -654,12 +770,12 @@ void Screen::updateEffectiveRendition()
         _effectiveBackground = _currentBackground;
     }
 
-    if ((_currentRendition & RE_BOLD) != 0) {
-        if ((_currentRendition & RE_FAINT) == 0) {
+    if ((_currentRendition.f.bold) != 0) {
+        if ((_currentRendition.f.faint) == 0) {
             _effectiveForeground.setIntensive();
         }
     } else {
-        if ((_currentRendition & RE_FAINT) != 0) {
+        if ((_currentRendition.f.faint) != 0) {
             _effectiveForeground.setFaint();
         }
     }
@@ -674,7 +790,7 @@ void Screen::copyFromHistory(Character *dest, int startLine, int count) const
     for (int line = startLine; line < startLine + count; ++line) {
         const int length = qMin(columns, _history->getLineLen(line));
         const int destLineOffset = (line - startLine) * columns;
-        const int lastColumn = (_history->getLineProperty(line) & LINE_DOUBLEWIDTH) ? columns / 2 : columns;
+        const int lastColumn = (_history->getLineProperty(line).flags.f.doublewidth) ? columns / 2 : columns;
 
         _history->getCells(line, 0, length, dest + destLineOffset);
 
@@ -688,7 +804,7 @@ void Screen::copyFromHistory(Character *dest, int startLine, int count) const
         if (_selBegin != -1) {
             for (int column = 0; column < lastColumn; ++column) {
                 if (isSelected(column, line)) {
-                    dest[destLineOffset + column].rendition |= RE_SELECTED;
+                    dest[destLineOffset + column].rendition.f.selected = 1;
                 }
             }
         }
@@ -705,7 +821,7 @@ void Screen::copyFromScreen(Character *dest, int startLine, int count) const
 
     for (int line = startLine; line < endLine; ++line) {
         const int destLineOffset = (line - startLine) * columns;
-        const int lastColumn = (line < (int)_lineProperties.size() && _lineProperties[line] & LINE_DOUBLEWIDTH) ? columns / 2 : columns;
+        const int lastColumn = (line < (int)_lineProperties.size() && _lineProperties[line].flags.f.doublewidth) ? columns / 2 : columns;
         const ImageLine srcLine = _screenLines.at(line);
         const int length = qMin(columns, srcLine.size());
 
@@ -720,7 +836,7 @@ void Screen::copyFromScreen(Character *dest, int startLine, int count) const
         if (_selBegin != -1) {
             for (int column = 0; column < lastColumn; ++column) {
                 if (isSelected(column, line + historyLines)) {
-                    dest[destLineOffset + column].rendition |= RE_SELECTED;
+                    dest[destLineOffset + column].rendition.f.selected = 1;
                 }
             }
         }
@@ -761,7 +877,12 @@ void Screen::getImage(Character *dest, int size, int startLine, int endLine) con
     // mark the character at the current cursor position
     int cursorIndex = loc(visX, _cuY + linesInHistoryBuffer);
     if (getMode(MODE_Cursor) && cursorIndex < _columns * mergedLines) {
-        dest[cursorIndex].rendition |= RE_CURSOR;
+        dest[cursorIndex].rendition.f.cursor = 1;
+    }
+    cursorIndex = loc(_selCuX, _selCuY - startLine + _history->getLines());
+
+    if (getMode(MODE_SelectCursor) && cursorIndex >= 0 && cursorIndex < _columns * mergedLines) {
+        dest[cursorIndex].rendition.f.cursor = 1;
     }
 }
 
@@ -795,7 +916,7 @@ QVector<LineProperty> Screen::getLineProperties(int startLine, int endLine) cons
 
 int Screen::getScreenLineColumns(const int line) const
 {
-    if (line < (int)_lineProperties.size() && _lineProperties.at(line) & LINE_DOUBLEWIDTH) {
+    if (line < (int)_lineProperties.size() && _lineProperties.at(line).flags.f.doublewidth) {
         return _columns / 2;
     }
 
@@ -837,6 +958,7 @@ void Screen::reset(bool softReset, bool preservePrompt)
     saveMode(MODE_Insert); // overstroke
 
     setMode(MODE_Cursor); // cursor visible
+    resetMode(MODE_SelectCursor);
 
     _topMargin = 0;
     _bottomMargin = _lines - 1;
@@ -920,11 +1042,12 @@ void Screen::initTabStops()
 void Screen::newLine()
 {
     if (getMode(MODE_NewLine)) {
-        _lineProperties[_cuY] = SetLineLength(_lineProperties[_cuY], _cuX);
+        _lineProperties[_cuY].length = _cuX;
         toStartOfLine();
     }
 
     index();
+    _lineProperties[_cuY].counter = commandCounter;
 }
 
 void Screen::checkSelection(int from, int to)
@@ -947,13 +1070,14 @@ void Screen::displayCharacter(uint c)
     // putting the cursor one right to the last column of the screen.
 
     int w = Character::width(c);
-
+    const QChar::Category category = QChar::category(c);
     if (w < 0) {
         // Non-printable character
         return;
-    } else if (w == 0) {
-        const QChar::Category category = QChar::category(c);
-        if (category != QChar::Mark_NonSpacing && category != QChar::Letter_Other && category != QChar::Other_Format) {
+    } else if (category == QChar::Mark_SpacingCombining || w == 0 || Character::emoji(c) || c == 0x20E3) {
+        bool emoji = Character::emoji(c);
+        if (category != QChar::Mark_SpacingCombining && category != QChar::Mark_NonSpacing && category != QChar::Letter_Other && category != QChar::Other_Format
+            && !emoji && c != 0x20E3) {
             return;
         }
         // Find previous "real character" to try to combine with
@@ -963,7 +1087,7 @@ void Screen::displayCharacter(uint c)
         do {
             if (charToCombineWithX > 0) {
                 --charToCombineWithX;
-            } else if (charToCombineWithY > 0 && _lineProperties.at(charToCombineWithY - 1) & LINE_WRAPPED) { // Try previous line
+            } else if (charToCombineWithY > 0 && _lineProperties.at(charToCombineWithY - 1).flags.f.wrapped) { // Try previous line
                 --charToCombineWithY;
                 charToCombineWithX = _screenLines.at(charToCombineWithY).length() - 1;
             } else {
@@ -980,6 +1104,9 @@ void Screen::displayCharacter(uint c)
         } while (_screenLines.at(charToCombineWithY).at(charToCombineWithX).isRightHalfOfDoubleWide());
 
         if (!previousChar) {
+            if (emoji) {
+                goto notcombine;
+            }
             if (!Hangul::isHangul(c)) {
                 return;
             } else {
@@ -990,24 +1117,88 @@ void Screen::displayCharacter(uint c)
 
         Character &currentChar = _screenLines[charToCombineWithY][charToCombineWithX];
 
+        if (c == 0x20E3) {
+            // Combining Enclosing Keycap - only combines with presentation mode #,*,0-9
+            if ((currentChar.character != 0x23 && currentChar.character != 0x2A && (currentChar.character < '0' || currentChar.character > '9'))
+                || (currentChar.flags & EF_EMOJI_REPRESENTATION) == 0) {
+                // Is this the right thing TODO?
+                return;
+            }
+        }
+        if (c == 0xFE0F) {
+            // Emoji presentation - should not be included
+            currentChar.flags |= EF_EMOJI_REPRESENTATION;
+            return;
+        }
+        if (c == 0x200D) {
+            // Zero width joiner
+            currentChar.flags |= EF_EMOJI_REPRESENTATION;
+        }
+        if (c >= 0xE0020 && c <= 0xE007F) {
+            // Tags - used for some flags
+            currentChar.flags |= EF_EMOJI_REPRESENTATION;
+        }
+
+        if (c >= 0x1f3fb && c <= 0x1f3ff) {
+            // Emoji modifier Fitzpatrick - changes skin color
+            uint currentUcs4 = currentChar.character;
+            if (currentChar.rendition.f.extended == 1) {
+                ushort extendedCharLength;
+                const uint *oldChars = ExtendedCharTable::instance.lookupExtendedChar(currentChar.character, extendedCharLength);
+                currentUcs4 = oldChars[extendedCharLength - 1];
+            }
+            if (currentUcs4 < 0x261d || (currentUcs4 > 0x270d && currentUcs4 < 0x1efff) || currentUcs4 > 0x1faff) {
+                goto notcombine;
+            }
+            currentChar.flags |= EF_EMOJI_REPRESENTATION;
+        } else if (c >= 0x1f1e6 && c <= 0x1f1ff) {
+            // Regional indicators - flag components
+            if (currentChar.rendition.f.extended == 1 || currentChar.character < 0x1f1e6 || currentChar.character > 0x1f1ff) {
+                goto notcombine;
+            }
+            currentChar.flags |= EF_EMOJI_REPRESENTATION;
+        } else if (emoji) {
+            if (currentChar.rendition.f.extended == 0) {
+                goto notcombine;
+            }
+            ushort extendedCharLength;
+            const uint *oldChars = ExtendedCharTable::instance.lookupExtendedChar(currentChar.character, extendedCharLength);
+            if (oldChars[extendedCharLength - 1] != 0x200d) {
+                goto notcombine;
+            }
+        }
+
         if (Hangul::isHangul(c) && !Hangul::combinesWith(currentChar, c)) {
             w = 2;
             goto notcombine;
         }
 
-        if ((currentChar.rendition & RE_EXTENDED_CHAR) == 0) {
+        if (currentChar.rendition.f.extended == 0) {
             const uint chars[2] = {currentChar.character, c};
-            currentChar.rendition |= RE_EXTENDED_CHAR;
+            currentChar.rendition.f.extended = 1;
             auto extChars = [this]() {
                 return usedExtendedChars();
             };
             currentChar.character = ExtendedCharTable::instance.createExtendedChar(chars, 2, extChars);
+            if (category == QChar::Mark_SpacingCombining) {
+                // ensure current line vector has enough elements
+                if (_screenLines[_cuY].size() < _cuX + w) {
+                    _screenLines[_cuY].resize(_cuX + w);
+                }
+                Character &ch = _screenLines[_cuY][_cuX];
+                ch.setRightHalfOfDoubleWide();
+                ch.foregroundColor = _effectiveForeground;
+                ch.backgroundColor = _effectiveBackground;
+                ch.rendition = _effectiveRendition;
+                ch.flags = setRepl(EF_UNREAL, _replMode);
+                _cuX += 1;
+            }
         } else {
             ushort extendedCharLength;
             const uint *oldChars = ExtendedCharTable::instance.lookupExtendedChar(currentChar.character, extendedCharLength);
+            Q_ASSERT(extendedCharLength > 1);
             Q_ASSERT(oldChars);
-            if (((oldChars) != nullptr) && extendedCharLength < 8) {
-                Q_ASSERT(extendedCharLength > 1);
+            if (((oldChars) != nullptr) && extendedCharLength < 10) {
                 Q_ASSERT(extendedCharLength < 65535); // redundant due to above check
                 auto chars = std::make_unique<uint[]>(extendedCharLength + 1);
                 std::copy_n(oldChars, extendedCharLength, chars.get());
@@ -1024,7 +1215,7 @@ void Screen::displayCharacter(uint c)
 notcombine:
     if (_cuX + w > getScreenLineColumns(_cuY)) {
         if (getMode(MODE_Wrap)) {
-            _lineProperties[_cuY] = static_cast<LineProperty>(_lineProperties.at(_cuY) | LINE_WRAPPED);
+            _lineProperties[_cuY].flags.f.wrapped = 1;
             nextLine();
         } else {
             _cuX = qMax(getScreenLineColumns(_cuY) - w, 0);
@@ -1051,7 +1242,19 @@ notcombine:
     currentChar.foregroundColor = _effectiveForeground;
     currentChar.backgroundColor = _effectiveBackground;
     currentChar.rendition = _effectiveRendition;
-    currentChar.flags = setRepl(EF_REAL, _replMode);
+    currentChar.flags = setRepl(EF_REAL, _replMode) | SetULColor(0, _currentULColor);
+    if (Character::emojiPresentation(c)) {
+        currentChar.flags |= EF_EMOJI_REPRESENTATION;
+    }
+    if (c <= '~' && c > ' ') {
+        currentChar.flags |= EF_ASCII_WORD;
+    }
+    if (c >= 0x900
+        && (c <= 0x109f || (c >= 0x1700 && c <= 0x18af) || (c >= 0x1900 && c <= 0x1aaf) || (c >= 0x1b00 && c <= 0x1c4f) || (c >= 0xa800 && c <= 0xa82f)
+            || (c >= 0xa840 && c <= 0xa95f) || (c >= 0xa980 && c <= 0xaaff) || (c >= 0xabc0 && c <= 0xabff) || (c >= 0x10a00 && c <= 0x10a5f)
+            || (c >= 0x11000 && c <= 0x11fff))) {
+        currentChar.flags |= EF_BRAHMIC_WORD;
+    }
 
     _lastDrawnChar = c;
 
@@ -1077,8 +1280,8 @@ notcombine:
     if (_replMode != REPL_None && std::make_pair(_cuY, _cuX) >= _replModeEnd) {
         _replModeEnd = std::make_pair(_cuY, _cuX);
     }
-    if (LineLength(_lineProperties[_cuY]) < _cuX) {
-        _lineProperties[_cuY] = SetLineLength(_lineProperties[_cuY], _cuX);
+    if (_lineProperties[_cuY].length < _cuX) {
+        _lineProperties[_cuY].length = _cuX;
     }
 
     if (_escapeSequenceUrlExtractor) {
@@ -1142,6 +1345,7 @@ void Screen::scrollUp(int from, int n)
     if (_hasGraphics) {
         scrollPlacements(n);
     }
+    _selCuY = qMax(_selCuY - n, -_history->getLines());
     if (_replMode != REPL_None) {
         if (_replModeStart.first > 0) {
             _replModeStart = std::make_pair(_replModeStart.first - 1, _replModeStart.second);
@@ -1251,16 +1455,16 @@ void Screen::clearImage(int loca, int loce, char c, bool resetLineRendition)
         const int startCol = (y == topLine) ? loca % _columns : 0;
 
         if (endCol < _columns - 1 || startCol > 0) {
-            _lineProperties[y] &= ~LINE_WRAPPED;
-            if (LineLength(_lineProperties[y]) < endCol && LineLength(_lineProperties[y]) > startCol) {
-                _lineProperties[y] = SetLineLength(_lineProperties[y], startCol);
+            _lineProperties[y].flags.f.wrapped = 0;
+            if (_lineProperties[y].length < endCol && _lineProperties[y].length > startCol) {
+                _lineProperties[y].length = startCol;
             }
         } else {
             if (resetLineRendition) {
-                _lineProperties[y] = LINE_DEFAULT;
+                _lineProperties[y] = LineProperty();
             }
             {
-                _lineProperties[y] &= ~(LINE_WRAPPED | LINE_PROMPT_START | LINE_INPUT_START | LINE_OUTPUT_START);
+                _lineProperties[y].flags.all &= ~(LINE_WRAPPED | LINE_PROMPT_START | LINE_INPUT_START | LINE_OUTPUT_START);
             }
         }
 
@@ -1416,13 +1620,19 @@ void Screen::clearEntireLine()
 
 void Screen::setRendition(RenditionFlags rendition)
 {
-    _currentRendition |= rendition;
+    _currentRendition.all |= rendition;
+    updateEffectiveRendition();
+}
+
+void Screen::setUnderlineType(int type)
+{
+    _currentRendition.f.underline = type;
     updateEffectiveRendition();
 }
 
 void Screen::resetRendition(RenditionFlags rendition)
 {
-    _currentRendition &= ~rendition;
+    _currentRendition.all &= ~rendition;
     updateEffectiveRendition();
 }
 
@@ -1430,7 +1640,8 @@ void Screen::setDefaultRendition()
 {
     setForeColor(COLOR_SPACE_DEFAULT, DEFAULT_FORE_COLOR);
     setBackColor(COLOR_SPACE_DEFAULT, DEFAULT_BACK_COLOR);
-    _currentRendition = DEFAULT_RENDITION;
+    _currentULColor = 0;
+    _currentRendition = {DEFAULT_RENDITION};
     updateEffectiveRendition();
 }
 
@@ -1456,6 +1667,31 @@ void Screen::setBackColor(int space, int color)
     }
 }
 
+void Screen::setULColor(int space, int color)
+{
+    CharacterColor col(quint8(space), color);
+    if (col.isValid()) {
+        int end = _ulColorQueueEnd;
+        if (end < _ulColorQueueStart) {
+            end += 15;
+        }
+        for (int i = _ulColorQueueStart; i < end; i++) {
+            if (col == _ulColors[i % 15]) {
+                _currentULColor = i % 15 + 1;
+                return;
+            }
+        }
+        _ulColors[_ulColorQueueEnd] = col;
+        _currentULColor = _ulColorQueueEnd + 1;
+        _ulColorQueueEnd = (_ulColorQueueEnd + 1) % 15;
+        if (_ulColorQueueEnd == _ulColorQueueStart) {
+            _ulColorQueueStart = (_ulColorQueueStart + 1) % 15;
+        }
+    } else {
+        _currentULColor = 0;
+    }
+}
+
 void Screen::clearSelection()
 {
     _selBottomRight = -1;
@@ -1531,40 +1767,62 @@ void Screen::setSelectionEnd(const int x, const int y, const bool trimTrailingWh
 
         _selTopLeft = loc(qMin(topColumn, bottomColumn), topRow);
         _selBottomRight = loc(qMax(topColumn, bottomColumn), bottomRow);
+        return;
+    }
+    // Extend the selection to the rightmost column if beyond the last character in the line
+    const int bottomRow = _selBottomRight / _columns;
+    const int bottomColumn = _selBottomRight % _columns;
+
+    bool beyondLastColumn = true;
+    if (bottomRow < _history->getLines()) {
+        ImageLine histLine;
+        const int histLineLen = _history->getLineLen(bottomRow);
+        histLine.resize(histLineLen);
+
+        _history->getCells(bottomRow, 0, histLineLen, histLine.data());
+
+        for (int j = bottomColumn; j < histLineLen; j++) {
+            if ((histLine.at(j).flags & EF_REAL) != 0 && (!trimTrailingWhitespace || !QChar(histLine.at(j).character).isSpace())) {
+                beyondLastColumn = false;
+            }
+        }
     } else {
-        // Extend the selection to the rightmost column if beyond the last character in the line
-        const int bottomRow = _selBottomRight / _columns;
-        const int bottomColumn = _selBottomRight % _columns;
+        size_t line = bottomRow - _history->getLines();
+        const int lastColumn = (line < _lineProperties.size() && _lineProperties[line].flags.f.doublewidth) ? _columns / 2 : _columns;
+        const auto *data = _screenLines[line].data();
 
-        bool beyondLastColumn = true;
-        if (bottomRow < _history->getLines()) {
-            ImageLine histLine;
-            const int histLineLen = _history->getLineLen(bottomRow);
-            histLine.resize(histLineLen);
+        // This should never happen, but it's happening. this is just to gather some information
+        // about the crash.
+        // Do not let this code go to a release.
+        if (_screenLines.size() < line) {
+            QFile konsoleInfo(QStringLiteral("~/konsole_info_crash_array_out_of_bounds.txt"));
+            konsoleInfo.open(QIODevice::WriteOnly);
+            QTextStream messages(&konsoleInfo);
 
-            _history->getCells(bottomRow, 0, histLineLen, histLine.data());
-
-            for (int j = bottomColumn; j < histLineLen; j++) {
-                if ((histLine.at(j).flags & EF_REAL) != 0 && (!trimTrailingWhitespace || !QChar(histLine.at(j).character).isSpace())) {
-                    beyondLastColumn = false;
-                }
-            }
-        } else {
-            const size_t line = bottomRow - _history->getLines();
-            const int lastColumn = (line < _lineProperties.size() && _lineProperties[line] & LINE_DOUBLEWIDTH) ? _columns / 2 : _columns;
-            const auto *data = _screenLines[line].data();
-            const int length = _screenLines.at(line).count();
-
-            for (int k = bottomColumn; k < lastColumn && k < length; k++) {
-                if ((data[k].flags & EF_REAL) != 0 && (!trimTrailingWhitespace || !QChar(data[k].character).isSpace())) {
-                    beyondLastColumn = false;
-                }
-            }
+            messages << "_selBegin" << _selBegin << "\n";
+            messages << "endPos" << endPos << "\n";
+            messages << "_selBottomRight" << _selBottomRight << "\n";
+            messages << "bottomRow Calculation: (_selBottomRight / _columns) = " << _selBottomRight << "/" << _columns << "\n";
+            messages << "line Calculation: (bottomRow - _history->getLines()) = " << bottomRow << "-" << _history->getLines() << "\n";
+            messages << "_screenLines.count()" << _screenLines.size() << "\n";
+            messages << "line" << line << "\n";
         }
 
-        if (beyondLastColumn) {
-            _selBottomRight = loc(_columns - 1, bottomRow);
+        // HACK: do not crash.
+        if (_screenLines.size() < line) {
+            line = _screenLines.size() - 1;
         }
+        const int length = _screenLines.at(line).count();
+
+        for (int k = bottomColumn; k < lastColumn && k < length; k++) {
+            if ((data[k].flags & EF_REAL) != 0 && (!trimTrailingWhitespace || !QChar(data[k].character).isSpace())) {
+                beyondLastColumn = false;
+            }
+        }
+    }
+
+    if (beyondLastColumn) {
+        _selBottomRight = loc(_columns - 1, bottomRow);
     }
 }
 
@@ -1608,7 +1866,7 @@ void Screen::selectReplContigious(const int x, const int y)
     if (row < _history->getLines()) {
         col = std::min(col, _history->getLineLen(row) - 1);
     } else {
-        col = std::min(col, _screenLines[row - _history->getLines()].size() - 1);
+        col = std::min(col, int(_screenLines[row - _history->getLines()].size()) - 1);
     }
     while (col > 0 && (getCharacter(col, row).flags & EF_REPL) == EF_REPL_NONE) {
         col--;
@@ -1757,7 +2015,7 @@ void Screen::writeToStream(TerminalCharacterDecoder *decoder, int startIndex, in
         // the text on a line.
         if (y == bottom && copied < count && !options.testFlag(TrimTrailingWhitespace)) {
             Character newLineChar('\n');
-            decoder->decodeLine(&newLineChar, 1, 0);
+            decoder->decodeLine(&newLineChar, 1, LineProperty());
         }
     }
 }
@@ -1802,7 +2060,7 @@ int Screen::copyLineToStream(int line,
     // ensure that this method, can append space or 'eol' character to
     // the selection
     Character *characterBuffer = getCharacterBuffer((count > -1 ? count : lineLength - start) + 1);
-    LineProperty currentLineProperties = 0;
+    LineProperty currentLineProperties = LineProperty();
 
     // determine if the line is in the history buffer or the screen image
     if (line < _history->getLines()) {
@@ -1825,16 +2083,16 @@ int Screen::copyLineToStream(int line,
 
         // Exclude trailing empty cells from count and don't bother processing them further.
         // See the comment on the similar case for screen lines for an explanation.
-        while (count > 0 && (characterBuffer[start + count - 1].flags & EF_REAL) == 0) {
+        while (count > 0 && (characterBuffer[count - 1].flags & EF_REAL) == 0) {
             count--;
         }
 
         if (_history->isWrappedLine(line)) {
-            currentLineProperties |= LINE_WRAPPED;
+            currentLineProperties.flags.f.wrapped = 1;
         } else {
             if (options.testFlag(TrimTrailingWhitespace)) {
                 // ignore trailing white space at the end of the line
-                while (count > 0 && QChar(characterBuffer[start + count - 1].character).isSpace()) {
+                while (count > 0 && QChar(characterBuffer[count - 1].character).isSpace()) {
                     count--;
                 }
             }
@@ -1866,7 +2124,7 @@ int Screen::copyLineToStream(int line,
         }
 
         // Don't remove end spaces in lines that wrap
-        if (options.testFlag(TrimTrailingWhitespace) && ((_lineProperties.at(screenLine) & LINE_WRAPPED) == 0)) {
+        if (options.testFlag(TrimTrailingWhitespace) && ((_lineProperties.at(screenLine).flags.f.wrapped) == 0)) {
             // ignore trailing white space at the end of the line
             while (length > 0 && QChar(data[length - 1].character).isSpace()) {
                 length--;
@@ -1883,14 +2141,14 @@ int Screen::copyLineToStream(int line,
         count = qBound(0, count, length - start);
 
         Q_ASSERT((size_t)screenLine < _lineProperties.size());
-        currentLineProperties |= _lineProperties[screenLine];
+        currentLineProperties = _lineProperties[screenLine];
     }
 
     if (appendNewLine) {
         // When users ask not to preserve the linebreaks, they usually mean:
         // `treat LINEBREAK as SPACE, thus joining multiple _lines into
         // single line in the same way as 'J' does in VIM.`
-        const bool isLineWrapped = (currentLineProperties & LINE_WRAPPED) != 0;
+        const bool isLineWrapped = (currentLineProperties.flags.f.wrapped) != 0;
         if (isBlockSelectionMode || !isLineWrapped) {
             characterBuffer[count] = options.testFlag(PreserveLineBreaks) ? Character('\n') : Character(' ');
             ++count;
@@ -1920,7 +2178,6 @@ int Screen::copyLineToStream(int line,
         int p = 0;
         for (int i = 0; i < count; i++) {
             Character c = characterBuffer[spacesCount + i];
-            // fprintf(stderr, "copy %i   %i %i\n", i, c.character, c.flags);
             if (((options & ExcludePrompt) != 0 && (c.flags & EF_REPL) == EF_REPL_PROMPT)
                 || ((options & ExcludeInput) != 0 && (c.flags & EF_REPL) == EF_REPL_INPUT)
                 || ((options & ExcludeOutput) != 0 && (c.flags & EF_REPL) == EF_REPL_OUTPUT)) {
@@ -2069,12 +2326,12 @@ const HistoryType &Screen::getScroll() const
     return _history->getType();
 }
 
-void Screen::setLineProperty(LineProperty property, bool enable)
+void Screen::setLineProperty(quint16 property, bool enable)
 {
     if (enable) {
-        _lineProperties[_cuY] = static_cast<LineProperty>(_lineProperties.at(_cuY) | property);
+        _lineProperties[_cuY].flags.all |= property;
     } else {
-        _lineProperties[_cuY] = static_cast<LineProperty>(_lineProperties.at(_cuY) & ~property);
+        _lineProperties[_cuY].flags.all &= ~property;
     }
 }
 
@@ -2084,6 +2341,8 @@ void Screen::setReplMode(int mode)
         if (_replMode == REPL_OUTPUT) {
             _replLastOutputStart = _replModeStart;
             _replLastOutputEnd = _replModeEnd;
+        } else if (_replMode == REPL_PROMPT) {
+            _lineProperties[_cuY].counter = ++commandCounter;
         }
         _replMode = mode;
         _replModeStart = std::make_pair(_cuY, _cuX);
@@ -2096,6 +2355,26 @@ void Screen::setReplMode(int mode)
     }
 }
 
+void Screen::setExitCode(int /*exitCode*/)
+{
+    int y = _cuY - 1;
+    while (y >= 0) {
+        _lineProperties[y].flags.f.error = 1;
+        if (_lineProperties[y].flags.f.prompt_start) {
+            return;
+        }
+        y--;
+    }
+    while (y > -_history->getLines()) {
+        LineProperty prop = _history->getLineProperty(y + _history->getLines());
+        prop.flags.f.error = 1;
+        _history->setLineProperty(y + _history->getLines(), prop);
+        if (prop.flags.f.prompt_start) {
+            return;
+        }
+        y--;
+    }
+}
 void Screen::fillWithDefaultChar(Character *dest, int count)
 {
     std::fill_n(dest, count, Screen::DefaultChar);
